@@ -174,12 +174,16 @@ static gboolean			srl_idle_installed	= FALSE;
 static GQueue			*srl_event_queue	= NULL;
 static AccessibleEventListener 	*srl_event_listeners[G_N_ELEMENTS (srl_events_type_name)];
 static Accessible		*srl_last_focus		= NULL;
+static Accessible		*srl_watched_acc	= NULL;
+static Accessible 		*srl_last_table 	= NULL;
 /*FIXME:next varaible should be in SRObject.c file */
 Accessible 			*srl_last_edit		= NULL;
 
 extern gboolean sro_set_difference (SRObject *obj, gchar *difference);
+extern gboolean sro_set_name (SRObject *obj, gchar *name);
 extern gboolean sro_get_from_accessible_event (Accessible *acc, gchar *reason, 
 						SRObject **obj);
+extern Accessible* sro_get_acc (SRObject *obj);
 #define sro_get_from_acc	sro_get_from_accessible_event
 
 #ifdef SRL_TIMING
@@ -708,6 +712,7 @@ srl_is_focus_event (SRLEvent *event)
 	    event->type == SRL_EVENT_TEXT_CHANGED_DELETE||
 	    event->type == SRL_EVENT_TEXT_SELECTION_CHANGED||
 	    event->type == SRL_EVENT_VALUE_CHANGED	||
+	    event->type == SRL_EVENT_NAME_CHANGED	||
 	    event->type == SRL_EVENT_STATE_CHECKED	||
 	    event->type == SRL_EVENT_STATE_SELECTED	||
 	    event->type == SRL_EVENT_STATE_EXPANDED	||
@@ -764,13 +769,70 @@ srl_notify_all_clients (SREvent *event,
     return TRUE;
 }
 
+static gboolean
+srl_set_last_table (Accessible *acc)
+{
+    if (srl_last_table)
+	Accessible_unref (srl_last_table);
+    srl_last_table = acc;
+    if (srl_last_table)
+	Accessible_ref (srl_last_table);
+    return TRUE;
+}
+
+static gboolean
+srl_table_same_line (Accessible *acc)
+{
+    static gint last_row = -1;
+    Accessible *parent;
+    gboolean rv = FALSE;
+
+    srl_assert (acc);
+
+    parent = Accessible_getParent (acc);
+    if (parent)
+    {
+	AccessibleTable *table= Accessible_getTable (parent);
+	if (table)
+	{
+	    gint row = AccessibleTable_getRowAtIndex (table, 
+				    Accessible_getIndexInParent (acc));
+	    if (row == last_row && parent == srl_last_table)
+	        rv = TRUE;
+	    last_row = row;
+	    AccessibleTable_unref (table);
+	}
+	Accessible_unref (parent);
+    }
+    return rv;
+}
+
+
+static gboolean
+srl_table_same_header (Accessible *acc)
+{
+    Accessible *parent;
+    gboolean rv = FALSE;
+
+    srl_assert (acc);
+
+    parent = Accessible_getParent (acc);
+    if (parent)
+    {
+	if (parent == srl_last_table)
+	    rv = TRUE;
+	Accessible_unref (parent);
+    }
+    return rv;
+}
+
 static gboolean
 srl_notify_clients_obj (SRLEvent *event,
 			SREventType type)
 {
     SREvent *ev;
     SRObject *obj;
-
+    
     srl_assert (event);
 
     ev = sre_new ();
@@ -780,6 +842,32 @@ srl_notify_clients_obj (SRLEvent *event,
 				    (gchar*)srle_get_reason (event), &obj))
 	obj = NULL;
 
+    if (obj)
+    {
+	static SRObjectRoles last_role = SR_ROLE_UNKNOWN; /* to report all headers
+		in case of click from a table cell to a column header */
+	SRObjectRoles role;
+	sro_get_role (obj, &role, SR_INDEX_CONTAINER);
+	if (srle_has_type (event, SRL_EVENT_FOCUS))
+	    if ((role == SR_ROLE_TABLE_LINE && 
+		    srl_table_same_line ((Accessible*)srle_get_acc (event))) ||
+	        (role == SR_ROLE_TABLE_COLUMNS_HEADER && role == last_role &&
+	    	    srl_table_same_header ((Accessible*)srle_get_acc (event))))
+		{
+		    sro_release_reference (obj);
+		    if (!sro_get_from_acc ((Accessible*)srle_get_acc (event),
+				SRL_EVENT_NAME_SELECTION_CHANGED, &obj))
+			obj = NULL;
+		}
+	if (role == SR_ROLE_TABLE_LINE || role == SR_ROLE_TABLE_COLUMNS_HEADER)
+	{
+	    Accessible *parent = Accessible_getParent ((Accessible*)srle_get_acc (event));
+	    srl_set_last_table (parent);
+	    Accessible_unref (parent);
+	    last_role = role;
+	}
+    }
+
     if (obj)
     {
 	ev->type = type;
@@ -788,13 +876,19 @@ srl_notify_clients_obj (SRLEvent *event,
 	if (srle_has_type (event, SRL_EVENT_TEXT_CHANGED_INSERT) ||
 	    srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE))
 	{
-/*FIXME: AccessibleTextSelectionChangedEvent_getSelectionString */
-/*FIXME: AccessibleWindowEvent_getTitleString */
 	    gchar *diff = AccessibleTextChangedEvent_getChangeString (event->acc_ev);
 	    if (diff)
 		sro_set_difference (obj, diff);
 	    SPI_freeString (diff);
 	}
+	if (srl_is_window_event (event))
+	{
+	    gchar *name = AccessibleWindowEvent_getTitleString (event->acc_ev);
+	    if (name)
+		sro_set_name (obj, name);
+	    SPI_freeString (name);
+	}
+/*FIXME: AccessibleTextSelectionChangedEvent_getSelectionString */
 	srl_log_gnopernicus_event_user_obj (event, ev);
 	srl_notify_all_clients (ev, 0);
     }
@@ -1067,7 +1161,7 @@ srle_change_type (SRLEvent *event)
 	if (srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
 	    event->type = SRL_EVENT_FOCUS;
     }
-
+    
     return TRUE;
 }
 
@@ -1205,7 +1299,7 @@ srl_text_event_is_reported (SRLEvent *event,
 	rv = TRUE;
 
     if (srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON))
-        rv = srl_info_is_value_changed (info, info2);
+        rv = !srl_info_is_value_changed (info, info2);
 
     /* when a text-changed event for child of combos is emitted,
        if there is a selection,then this change take place 
@@ -1287,23 +1381,35 @@ srl_selection_event_is_reported (SRLEvent *event)
     return rv;
 }
 
+static gboolean
+srle_is_for_watched_acc (SRLEvent *event)
+{
+    srl_assert (event);
+    
+    return srl_watched_acc == event->acc_ev->source;
+}
+
+
 static gboolean
 srl_event_is_reported (SRLEvent *event)
 {
     gboolean rv = FALSE;
-    static SRLLastInfo last_info;
+    static SRLLastInfo last_info_focus, last_info_watched;
+    SRLLastInfo *last_info = NULL;
 
     srl_assert (event);
+    last_info = srle_is_for_focused_acc (event) ? &last_info_focus : &last_info_watched;
 
     if (srl_is_window_event (event) 	||
 	srl_is_mouse_event (event)	||
 	srl_is_tooltip_event (event))
 	rv = TRUE;
     else if (srle_has_type (event, SRL_EVENT_FOCUS))
-	rv = TRUE;
+	rv = srle_acc_has_state (event, SPI_STATE_FOCUSED);
     else if (srle_has_type (event, SRL_EVENT_FOCUS1))
 	rv = TRUE;
-    else if (srle_is_for_focused_acc (event))
+    else if (srle_is_for_focused_acc (event) ||
+	     srle_is_for_watched_acc (event))
     {
 	if (srle_has_type (event, SRL_EVENT_STATE_EXPANDED)) 
 	    rv = srle_acc_has_role (event, SPI_ROLE_TABLE_CELL);
@@ -1321,30 +1427,35 @@ srl_event_is_reported (SRLEvent *event)
 		 srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE) ||
 		 srle_has_type (event, SRL_EVENT_TEXT_CARET_MOVED)    ||
 		 srle_has_type (event, SRL_EVENT_TEXT_SELECTION_CHANGED))
-	    rv = srl_text_event_is_reported (event, last_info);
+	    rv = srl_text_event_is_reported (event, *last_info);
 	else if (srle_has_type (event, SRL_EVENT_SELECTION_CHANGED))
 	    rv = srl_selection_event_is_reported (event);
+	else if (srle_has_type (event, SRL_EVENT_NAME_CHANGED))
+	    rv = TRUE;
 	else if (srle_has_type (event, SRL_EVENT_VALUE_CHANGED))
-	    rv = srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON);
+	    rv = srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON) ||
+		    srle_acc_has_role (event, SPI_ROLE_SLIDER);
     }
 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	rv = FALSE;
 
-    if (rv)
+    if (rv && (srle_is_for_focused_acc (event) ||
+	      srle_is_for_watched_acc (event)))
     {
-	srl_set_info (event->acc_ev->source, &last_info);
+	srl_assert (last_info);
+	srl_set_info (event->acc_ev->source, last_info);
 
-	/* for text-changed:delete event information from object contains
-	   values from before event was fired, not the present */
-	if (srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE))
-	{
-	    last_info.text_caret_offset     = event->acc_ev->detail1;
-	    last_info.text_selections_count = 0;
-	    last_info.text_char_count      -= event->acc_ev->detail2;
-	    last_info.text_crt_selection_length = 0;
-	}
+	if (srle_is_for_focused_acc (event))
+	    if (last_info != &last_info_focus)
+		last_info_focus = *last_info;
+	if (srle_is_for_watched_acc (event))
+	    if (last_info != &last_info_watched)
+		last_info_watched = *last_info;
     }
+    
+    if (srle_acc_has_role (event, SPI_ROLE_EDITBAR))
+	rv = TRUE;
 
     return rv;
 }
@@ -1370,7 +1481,6 @@ srl_process_event (SRLEvent *event)
     srl_timing_process_start ((gpointer) event->acc_ev);
     srle_set_acc (event);
     srle_change_type (event);
-    process = srl_event_is_reported (event);
 
     if (srle_has_type (event, SRL_EVENT_FOCUS))
 	srl_set_last_focus ((Accessible*)srle_get_acc (event));
@@ -1380,7 +1490,7 @@ srl_process_event (SRLEvent *event)
 /* FIXME: 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	report no focus if no other event */
-
+    process = srl_event_is_reported (event);
     if (process)
 	srl_report_event (event);
 
@@ -1453,7 +1563,9 @@ srl_init ()
     srl_last_edit	= NULL;
     srl_idle_need	= FALSE;
     srl_idle_installed	= FALSE;
-
+    srl_watched_acc	= NULL;
+    srl_last_table	= NULL;
+    
     srl_log_init ();
     srl_timing_init ();
 
@@ -1509,6 +1621,10 @@ srl_terminate ()
 	Accessible_unref (srl_last_focus);
     if (srl_last_edit)
         Accessible_unref (srl_last_edit);
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    if (srl_last_table)
+	Accessible_unref (srl_last_table);
     srl_log_terminate ();
     srl_timing_terminate ();
     srl_set_uninitialized ();
@@ -1612,3 +1728,29 @@ srl_mouse_button_up (gint button)
     
     return SPI_generateMouseEvent (-1,-1 , action) ? TRUE : FALSE;
 }		
+
+
+gboolean
+srl_set_watch_for_object (SRObject *obj)
+{
+    Accessible *acc;
+
+    srl_assert (obj);
+
+    srl_unwatch_all_objects ();
+    acc = sro_get_acc (obj);
+    Accessible_ref (acc);
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    srl_watched_acc = acc;
+
+    return TRUE;
+}
+
+void
+srl_unwatch_all_objects ()
+{
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    srl_watched_acc = NULL;
+}
